/*============================================================================
ACK-3D - Animation Construction Kit 3D

Original Author:
	Lary Myers
    5 Jacob Street
    Ballston Lake, NY 12019
    Compuserve: 72355,655

Portions by:
	John Gaines, Jr.
	3512 Pear Tree Court #23
	Silver Spring, MD 20906-2577
	Home phone: (301) 871-2635
	Compuserve: 73747,3557

File:	trig.c  --  Program to generate TRIG.DAT file.

Public Functions:

		int  main(int ArgCount, char ** ArgPtrs);
		void ShowHelp(void);
		int  AllocateTables(void);
		int  NullCheckFree(void ** PointerPointer);
		void FreeTables(void);
		void BuildTables(void);
		int  ReadOneTable(FILE * InputFile, long * Table);
		int  ReadTables(char *FileName);
		void PrintOneTable(char * TableName, long * Table);
		void PrintTables(void);
		int  WriteOneTable(FILE * OutputFile, long * Table);
		int  WriteTables(char *FileName);

Descr:	

  ***revision-history***
1 TRIG.C 18-Aug-93,22:13:32 Initial version of program to create TRIG.DAT.
2 TRIG.C 20-Aug-93,2:07:38 Lary's CS #.
  ***revision-history***
============================================================================*/
#include <stdio.h>
#include <math.h>
#include <alloc.h>
#include <string.h>
#include <limits.h>
#include <float.h>
#include "angle.h"

typedef	long double	Real_t;

long *	SinTableShift16;
long *	CosTableShift16;
long *	TanTableShift16;
long *	InvTanTableShift16;
long *	InvCosTableShift20;
long *	InvSinTableShift20;
long *	CosTableShift14;
long *	ViewCosTable;

#define	TABLE_SIZE	(INT_ANGLE_360 * sizeof(long))

const Real_t	Shift14 = 16384.0;
const Real_t	Shift16 = 65536.0;
const Real_t	Shift20 = 1048576.0;
const Real_t	AlmostZero = 0.0000001;

Real_t (*RoundFunc)(Real_t);

typedef enum {OpCreate, OpRead} Operation_t;
typedef enum {OutputText, OutputBinary} OutputFormat_t;

void 	ShowHelp(void);
int  	AllocateTables(void);
int  	NullCheckFree(void ** PointerPointer);
void 	FreeTables(void);
void 	BuildTables(void);
int  	ReadOneTable(FILE * InputFile, long * Table);
int  	ReadTables(char *FileName);
void 	PrintOneTable(char * TableName, long * Table);
void 	PrintTables(void);
int  	WriteOneTable(FILE * OutputFile, long * Table);
int  	WriteTables(char *FileName);
Real_t	RoundUp(Real_t Value);
Real_t	RoundDown(Real_t Value);
Real_t	RoundIn(Real_t Value);
Real_t	RoundOut(Real_t Value);

/*----------------------------------------------------------------------------
Func:	main  --  

Descr:	
----------------------------------------------------------------------------*/
int main(int ArgCount, char ** ArgPtrs)
	{
	char *			OptionCharPtr;
	Operation_t		Operation = OpCreate;
	OutputFormat_t	OutputFormat = OutputText;
	char			FileName[FILENAME_MAX];

	fputs( "ACK-3D TRIG.DAT Builder  v1.0  by John Gaines, Jr.\n\n", stderr);
	
	FileName[0] = '\0';
	RoundFunc	= RoundOut;

	for (--ArgCount, ++ArgPtrs; ArgCount; --ArgCount, ++ArgPtrs)
		if (**ArgPtrs == '-' || **ArgPtrs == '/')
			{
			OptionCharPtr = (*ArgPtrs) + 1;
			while (*OptionCharPtr)
				{
				switch (*OptionCharPtr)
					{
					case '?':
						ShowHelp();
						return 0;
					case 'r':
						Operation = OpRead;
						++OptionCharPtr;
						break;
					case 'b':
						OutputFormat = OutputBinary;
						++OptionCharPtr;
						break;
					case 'u':
						RoundFunc = RoundUp;
						++OptionCharPtr;
						break;
					case 'd':
						RoundFunc = RoundDown;
						++OptionCharPtr;
						break;
					case 'i':
						RoundFunc = RoundIn;
						++OptionCharPtr;
						break;
					case 'o':
						RoundFunc = RoundOut;
						++OptionCharPtr;
						break;
					default:
						printf( "Invalid option character '%c' in \"%s\" argument.\n",
							*OptionCharPtr, *ArgPtrs);
						++OptionCharPtr;
						break;
					}
				}
			}
		else if (FileName[0] == '\0')
			strcpy( FileName, *ArgPtrs);
		else
			{
			printf( "Too many filenames on command line: %s\n", *ArgPtrs);
			ShowHelp();
			return -1;
			}

	if (FileName[0] == '\0')
		strcpy( FileName, "TRIG.DAT");

	if (AllocateTables())
		{
		printf("Out of memory allocating tables.\n");
		return -1;
		}

	if (Operation == OpCreate)
		BuildTables();
	else if (ReadTables( FileName))
		{
		printf( "Error reading tables from %s.\n", FileName);
		return -1;
		}

	if (Operation == OpRead || OutputFormat == OutputText)
		PrintTables();
	else
		WriteTables( FileName);

	FreeTables();

	fputs( "Done.\n", stderr);

	return 0;
	}

/*----------------------------------------------------------------------------
Func:	ShowHelp  --  

Descr:	
----------------------------------------------------------------------------*/
void ShowHelp(void)
	{
	static char * HelpText[] =
		{
		"Syntax:\t TRIG [options] [filename]",
		"",
		"Options: -?\tShow this help screen.",
		"\t -b\tOutput binary file.",
		"\t -r\tRead binary input from filename.",
		"\t -u\tRound floating point values up when converting to longs.",
		"\t -d\tRound floating point values down when converting to longs.",
		"\t -i\tRound floating point values inward towards zero when converting",
			"\t\tto longs.",
		"\t -o\tRound floating point values outward away from zero when",
			"\t\tconverting to longs.",
		"",
		"If no arguments are given, default operation is to create the trig tables,",
		"rounding floating point values out when converting to longs, and print the",
		"tables out in readable C array format to stdout.",
		(char *)0
		};
	char **	TextPtr;

	for (TextPtr = HelpText; *TextPtr; ++TextPtr)
		puts(*TextPtr);
	}

/*----------------------------------------------------------------------------
Func:	AllocateTables  --  Allocate memory for all tables.

Descr:	
----------------------------------------------------------------------------*/
int AllocateTables(void)
	{
	fputs( "Allocating memory for tables.\n", stderr);

	SinTableShift16	   = calloc( INT_ANGLE_360, sizeof(long));
	CosTableShift16	   = calloc( INT_ANGLE_360, sizeof(long));
	TanTableShift16	   = calloc( INT_ANGLE_360, sizeof(long));
	InvTanTableShift16 = calloc( INT_ANGLE_360, sizeof(long));
	InvCosTableShift20 = calloc( INT_ANGLE_360, sizeof(long));
	InvSinTableShift20 = calloc( INT_ANGLE_360, sizeof(long));
	CosTableShift14	   = calloc( INT_ANGLE_360, sizeof(long));

	if (SinTableShift16	   == NULL ||
		CosTableShift16	   == NULL ||
		TanTableShift16	   == NULL ||
		InvTanTableShift16 == NULL ||
		InvCosTableShift20 == NULL ||
		InvSinTableShift20 == NULL ||
		CosTableShift14	   == NULL)
		return -1;

	return 0;
	}

/*----------------------------------------------------------------------------
Func:	NullCheckFree  --  Check a pointer NULL before free, then set to NULL.

Descr:	
----------------------------------------------------------------------------*/
int NullCheckFree(void ** PointerPointer)
	{
	if (PointerPointer != NULL)
		{
		if (PointerPointer != NULL)
			free( *PointerPointer);
		*PointerPointer = NULL;
		return 0;
		}
	return -1;
	}

/*----------------------------------------------------------------------------
Func:	FreeTables  --  Free all memory allocated by program.

Descr:	
----------------------------------------------------------------------------*/
void FreeTables(void)
	{
	fputs( "Freeing table memory.\n", stderr);

	NullCheckFree( &(void *)SinTableShift16	  );
	NullCheckFree( &(void *)CosTableShift16	  );
	NullCheckFree( &(void *)TanTableShift16	  );
	NullCheckFree( &(void *)InvTanTableShift16);
	NullCheckFree( &(void *)InvCosTableShift20);
	NullCheckFree( &(void *)InvSinTableShift20);
	NullCheckFree( &(void *)CosTableShift14	  );
	}

/*----------------------------------------------------------------------------
Func:	BuildTables  --  Build all trig tables into memory arrays.

Descr:	
----------------------------------------------------------------------------*/
void BuildTables(void)
	{
	Real_t	Radians;
	Real_t	SinValue;
	Real_t	CosValue;
	Real_t	TanValue;
	int		InternalDegree;

	fputs( "Building tables...\n", stderr);

	/* Stepping by internal degrees and calculating the equivalent radian
		angle at each iteration, we eliminate rounding errors which result
		from incrementing a floating point number at each iteration. */
	
	for (InternalDegree = 0; InternalDegree < INT_ANGLE_360; ++InternalDegree)
		{
		fprintf( stderr, "%d\r", InternalDegree);

		/* Radians = (2pi InternalDegree) / INT_ANGLE_360 reduces to: */

		Radians = (M_PI * InternalDegree) / (Real_t)INT_ANGLE_180;

		SinValue = sinl( Radians);
		CosValue = cosl( Radians);
		if (Radians == M_PI_2 || Radians == (M_PI + M_PI_2))
			TanValue = LONG_MAX / Shift16;
		else
			TanValue = tanl( Radians);

		SinTableShift16[InternalDegree]	= (long)(RoundFunc(SinValue * Shift16));
		CosTableShift16[InternalDegree]	= (long)(RoundFunc(CosValue * Shift16));
		TanTableShift16[InternalDegree]	= (long)(RoundFunc(TanValue * Shift16));
		CosTableShift14[InternalDegree]	= (long)(RoundFunc(CosValue * Shift14));

		/* When calculating the inverse trig functions, we have to check for
			0.0 values, in which case, we'll put LONG_MAX in our table. */

		if (TanValue <= AlmostZero && TanValue >= -AlmostZero)
			InvTanTableShift16[InternalDegree] = LONG_MAX;
		else
			InvTanTableShift16[InternalDegree] = (long)(RoundFunc(1.0 / TanValue * Shift16));

		if (CosValue <= AlmostZero && CosValue >= -AlmostZero)
			InvCosTableShift20[InternalDegree] = LONG_MAX;
		else
			InvCosTableShift20[InternalDegree] = (long)(RoundFunc(1.0 / CosValue * Shift20));

		if (SinValue <= AlmostZero && SinValue >= -AlmostZero)
			InvSinTableShift20[InternalDegree] = LONG_MAX;
		else
			InvSinTableShift20[InternalDegree] = (long)(RoundFunc(1.0 / SinValue * Shift20));
		}

	fputs( "Done building tables.\n", stderr);
	}

/*----------------------------------------------------------------------------
Func:	ReadOneTable  --  Read one trig table from binary file.

Descr:	
----------------------------------------------------------------------------*/
int ReadOneTable(FILE * InputFile, long * Table)
	{
	fputc( '.', stderr);

	if (fread(Table, sizeof(long), INT_ANGLE_360, InputFile) != INT_ANGLE_360)
		return -1;

	return 0;
	}

/*----------------------------------------------------------------------------
Func:	ReadTables  --  Read all trig tables from binary file.

Descr:	
----------------------------------------------------------------------------*/
int ReadTables(char *FileName)
	{
	FILE *	InputFile;
	
#define	READ(Table)	if (ReadOneTable(InputFile, Table))	\
							return -1;

	InputFile = fopen( FileName, "rb");
	if (InputFile == NULL)
		{
		return -1;
		}

	fputs( "Reading tables", stderr);

	READ( SinTableShift16);
	READ( CosTableShift16);
	READ( TanTableShift16);
	READ( InvTanTableShift16);
	READ( InvCosTableShift20);
	READ( InvSinTableShift20);
	READ( CosTableShift14);

	fputs( "\nDone reading tables.\n", stderr);

	return 0;
	}

/*----------------------------------------------------------------------------
Func:	PrintOneTable  --  Print a table in readable C array format to stdout.

Descr:	
----------------------------------------------------------------------------*/
void PrintOneTable(char * TableName, long * Table)
	{
	int		InternalDegree;
	int *	DegreeConvPtr;

	static int	DegreeConversionTable[] =
		{
		0,				0,
		INT_ANGLE_1,	1,
		INT_ANGLE_2,	2,
		INT_ANGLE_4,	4,
		INT_ANGLE_6,	6,
		INT_ANGLE_30,	30,
		INT_ANGLE_45,	45,
		INT_ANGLE_90,	90,
		INT_ANGLE_135,	135,
		INT_ANGLE_180,	180,
		INT_ANGLE_225,	225,
		INT_ANGLE_270,	270,
		INT_ANGLE_315,	315,
		INT_ANGLE_360,	360,
		-1
		};

	fputc( '.', stderr);

	printf( "long %s [] =\n\t{\n", TableName);

	DegreeConvPtr = DegreeConversionTable;

	for (InternalDegree = 0; InternalDegree < INT_ANGLE_360; ++InternalDegree)
		{
		printf( "\t%ld%s", Table[InternalDegree],
			(InternalDegree == INT_ANGLE_360 - 1) ? "" : ",");

		if (InternalDegree == *DegreeConvPtr)
			{
			printf( "\t/* %d */\n", DegreeConvPtr[1]);
			DegreeConvPtr += 2;
			}
		else
			putchar( '\n');
		}

	printf( "\t};\n\n");
	}

/*----------------------------------------------------------------------------
Func:	PrintTables  --  Print all trig tables to stdout.

Descr:	
----------------------------------------------------------------------------*/
void PrintTables(void)
	{
#define PRINT(tbl)	PrintOneTable( #tbl, tbl)

	fputs( "Printing tables", stderr);

	PRINT( SinTableShift16);
	PRINT( CosTableShift16);
	PRINT( TanTableShift16);
	PRINT( InvTanTableShift16);
	PRINT( InvCosTableShift20);
	PRINT( InvSinTableShift20);
	PRINT( CosTableShift14);

	fputs( "\nDone printing tables.\n", stderr);
	}

/*----------------------------------------------------------------------------
Func:	WriteOneTable  --  Write one trig table to binary file.

Descr:	
----------------------------------------------------------------------------*/
int WriteOneTable(FILE * OutputFile, long * Table)
	{
	fputc( '.', stderr);

	if (fwrite(Table, sizeof(long), INT_ANGLE_360, OutputFile) != INT_ANGLE_360)
		return -1;

	return 0;
	}

/*----------------------------------------------------------------------------
Func:	WriteTables  --  Write all trig tables to binary file.

Descr:	
----------------------------------------------------------------------------*/
int WriteTables(char *FileName)
	{
	FILE *	OutputFile;
	
#define	WRITE(Table)	if (WriteOneTable(OutputFile, Table))	\
							return -1;

	OutputFile = fopen( FileName, "wb");
	if (OutputFile == NULL)
		{
		return -1;
		}

	fputs( "Writing tables", stderr);

	WRITE( SinTableShift16);
	WRITE( CosTableShift16);
	WRITE( TanTableShift16);
	WRITE( InvTanTableShift16);
	WRITE( InvCosTableShift20);
	WRITE( InvSinTableShift20);
	WRITE( CosTableShift14);

	fputs( "\nDone writing tables.\n", stderr);

	return 0;
	}

/*----------------------------------------------------------------------------
Func:	RoundUp  --  Round numbers up towards more positive values.

Descr:	
----------------------------------------------------------------------------*/
Real_t RoundUp(Real_t Value)
	{
	return ceill(Value);
	}

/*----------------------------------------------------------------------------
Func:	RoundDown  --  Round numbers down towards more negative values.

Descr:	
----------------------------------------------------------------------------*/
Real_t RoundDown(Real_t Value)
	{
	return floorl(Value);
	}

/*----------------------------------------------------------------------------
Func:	RoundIn  --  Round numbers in towards zero.

Descr:	
----------------------------------------------------------------------------*/
Real_t RoundIn(Real_t Value)
	{
	if (Value > 0.0)
		return floorl(Value);
	else
		return ceill(Value);
	}

/*----------------------------------------------------------------------------
Func:	RoundOut  --  Round numbers out away from zero.

Descr:	
----------------------------------------------------------------------------*/
Real_t RoundOut(Real_t Value)
	{
	if (Value > 0.0)
		return ceill(Value);
	else
		return floorl(Value);
	}
